module Deus.Lens where

import Data.IORef
import Control.Applicative((<$>))

-- Fundamental concept: a Lens, aka functional reference

data Lens s v = Lens { unLens :: s -> (v,v->s) }

data IOLens s = IOLens { iolGet :: IO s , iolSet :: s -> IO () }

data MonadLens m s = MonadLens { mlGet :: m s , mlSet :: s -> m () }


lGet l s = case unLens l s of (v,_) -> v
lSet l s v = case unLens l s of (_,f) -> f v
lModify l f s = lSet l s (f (lGet l s))
l .= v = \s -> lSet l s v

iorefLens :: IORef s -> MonadLens IO s
iorefLens r = MonadLens { mlGet = readIORef r,
                          mlSet = \s -> writeIORef r s }
              
(>.>) :: (Monad m, Functor m) => MonadLens m s -> Lens s v -> MonadLens m v
ml >.> l = MonadLens { mlGet = lGet l <$> mlGet ml,
                       mlSet = \v -> do s <- mlGet ml
                                        mlSet ml (lSet l s v) }

(<.>) :: Lens s v -> Lens v w -> Lens s w
l <.> n = Lens (\s -> let v = lGet l s
                          w = lGet n v
                      in (w,
                          \w' -> lSet l s (lSet n v w')))
                       

withLens :: Monad m => MonadLens m s -> (s -> m (s,a))  -> m a
withLens (MonadLens { mlGet = get, mlSet = set }) act = do
  s <- get
  (s',a) <- act s
  set s'
  return a
  
withLens_ :: Monad m => MonadLens m s -> (s -> m s) -> m ()
withLens_ (MonadLens { mlGet = get, mlSet = set }) act = do
  s <- get
  s' <- act s
  set s'

